@@ -19,6 +19,9 @@ Maintainer:  Ryan Scott
 Portability: Template Haskell
 
 Template Haskell-related utilities.
+
+Note: this is an internal module, and as such, the API presented here is not
+guaranteed to be stable, even between minor releases of this library.
 -}
 module Data.Deriving.Internal where
 
@@ -38,6 +41,7 @@ import           Data.Map (Map)
 import           Data.Maybe
 import qualified Data.Set as Set
 import           Data.Set (Set)
+import qualified Data.Traversable as T
 
 import           Text.ParserCombinators.ReadPrec (ReadPrec)
 import qualified Text.Read.Lex as L
@@ -1022,9 +1026,10 @@ tag2ConExpr :: Type -> Q Exp
 tag2ConExpr ty = do
     iHash  <- newName "i#"
     ty' <- freshenType ty
+    let tvbs = avoidTypeInType $ freeVariablesWellScoped [ty']
     lam1E (conP iHashDataName [varP iHash]) $
         varE tagToEnumHashValName `appE` varE iHash
-            `sigE` return (ForallT (requiredTyVarsOfType ty') [] ty')
+            `sigE` return (ForallT tvbs [] ty')
             -- tagToEnum# is a hack, and won't typecheck unless it's in the
             -- immediate presence of a type ascription like so:
             --
@@ -1033,16 +1038,109 @@ tag2ConExpr ty = do
             -- We have to be careful when dealing with datatypes with type
             -- variables, since Template Haskell might reject the type variables
             -- we use for being out-of-scope. To avoid this, we explicitly
-            -- collect the type variable binders with requiredTyVarsOfType
-            -- and shove them into a ForallT. Also make sure to freshen the
-            -- bound type variables to avoid shadowed variable warnings when
-            -- -Wall is enabled.
-            --
-            -- Note that we do NOT collect the kind variable binders, since
-            -- a type signature like `forall k a. Foo (a :: k)` won't typecheck
-            -- unless TypeInType is enabled (i.e., if GHC 8.0 or later is being
-            -- used). Luckily, GHC seems to just accept kind variables, even if
-            -- they aren't actually bound in a ForallT.
+            -- collect the type variable binders and shove them into a ForallT
+            -- (using th-abstraction's quantifyType function). Also make sure
+            -- to freshen the bound type variables to avoid shadowed variable
+            -- warnings on old versions of GHC when -Wall is enabled.
+  where
+    -- Somewhat annoyingly, it's possible to generate code that requires
+    -- TypeInType (on old versions of GHC) for data types which didn't require
+    -- TypeInType to define. To avoid users having to turn on more language
+    -- extensions than is necessary, we filter out all kind variable binders.
+    -- Fortunately, old versions of GHC are quite alright with implicitly
+    -- quantifying kind variables, even in the type of a SigE.
+    --
+    -- This is rather tiresome, and while writing this function, I debated
+    -- whether to just forget about this nonsense and require users to
+    -- enable TypeInType to use the generated code. Alas, that would entail
+    -- a breaking change, so I decided against it at the time. If we ever make
+    -- some breaking change in the future, however, this would be at the top
+    -- of the list of things that I'd rip out.
+    avoidTypeInType :: [TyVarBndr] -> [TyVarBndr]
+#if __GLASGOW_HASKELL__ >= 806
+    avoidTypeInType = id
+#else
+    avoidTypeInType = go . map attachFreeKindVars
+      where
+        attachFreeKindVars :: TyVarBndr -> (TyVarBndr, [Name])
+        attachFreeKindVars tvb = (tvb, freeVariables (tvKind tvb))
+
+        go :: [(TyVarBndr, [Name])] -> [TyVarBndr]
+        go [] = []
+        go ((tvb, _):tvbsAndFVs)
+          | any (\(_, kindVars) -> tvName tvb `elem` kindVars) tvbsAndFVs
+          = tvbs'
+          | otherwise
+          = tvb:tvbs'
+          where
+            tvbs' = go tvbsAndFVs
+#endif
+
+primOrdFunTbl :: Map Name (Name, Name, Name, Name, Name)
+primOrdFunTbl = Map.fromList
+    [ (addrHashTypeName,   ( ltAddrHashValName
+                           , leAddrHashValName
+                           , eqAddrHashValName
+                           , geAddrHashValName
+                           , gtAddrHashValName
+                           ))
+    , (charHashTypeName,   ( ltCharHashValName
+                           , leCharHashValName
+                           , eqCharHashValName
+                           , geCharHashValName
+                           , gtCharHashValName
+                           ))
+    , (doubleHashTypeName, ( ltDoubleHashValName
+                           , leDoubleHashValName
+                           , eqDoubleHashValName
+                           , geDoubleHashValName
+                           , gtDoubleHashValName
+                           ))
+    , (floatHashTypeName,  ( ltFloatHashValName
+                           , leFloatHashValName
+                           , eqFloatHashValName
+                           , geFloatHashValName
+                           , gtFloatHashValName
+                           ))
+    , (intHashTypeName,    ( ltIntHashValName
+                           , leIntHashValName
+                           , eqIntHashValName
+                           , geIntHashValName
+                           , gtIntHashValName
+                           ))
+    , (wordHashTypeName,   ( ltWordHashValName
+                           , leWordHashValName
+                           , eqWordHashValName
+                           , geWordHashValName
+                           , gtWordHashValName
+                           ))
+#if MIN_VERSION_base(4,13,0)
+    , (int8HashTypeName,   ( ltInt8HashValName
+                           , leInt8HashValName
+                           , eqInt8HashValName
+                           , geInt8HashValName
+                           , gtInt8HashValName
+                           ))
+    , (int16HashTypeName,  ( ltInt16HashValName
+                           , leInt16HashValName
+                           , eqInt16HashValName
+                           , geInt16HashValName
+                           , gtInt16HashValName
+                           ))
+    , (word8HashTypeName,  ( ltWord8HashValName
+                           , leWord8HashValName
+                           , eqWord8HashValName
+                           , geWord8HashValName
+                           , gtWord8HashValName
+                           ))
+    , (word16HashTypeName, ( ltWord16HashValName
+                           , leWord16HashValName
+                           , eqWord16HashValName
+                           , geWord16HashValName
+                           , gtWord16HashValName
+                           ))
+#endif
+    ]
 
 removeClassApp :: Type -> Type
 removeClassApp (AppT _ t2) = t2
@@ -1058,13 +1156,10 @@ freshen :: Name -> Q Name
 freshen n = newName (nameBase n ++ "_'")
 
 freshenType :: Type -> Q Type
-freshenType (AppT t1 t2) = do t1' <- freshenType t1
-                              t2' <- freshenType t2
-                              return $ AppT t1' t2'
-freshenType (SigT t k)   = do t' <- freshenType t
-                              return $ SigT t' k
-freshenType (VarT n)     = fmap VarT $ freshen n
-freshenType t            = return t
+freshenType t =
+  do let xs = [(n, VarT `fmap` freshen n) | n <- freeVariables t]
+     subst <- T.sequence (Map.fromList xs)
+     return (applySubstitution subst t)
 
 -- | Gets all of the required type variable binders mentioned in a Type.
 requiredTyVarsOfType :: Type -> [TyVarBndr]
@@ -2009,3 +2104,89 @@ readSymField fieldName readVal = do
 readSymFieldValName :: Name
 readSymFieldValName = mkNameG_v derivingCompatPackageKey "Data.Deriving.Internal" "readSymField"
 #endif
+
+#if MIN_VERSION_base(4,13,0)
+eqInt8HashValName :: Name
+eqInt8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "eqInt8#"
+
+eqInt16HashValName :: Name
+eqInt16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "eqInt16#"
+
+eqWord8HashValName :: Name
+eqWord8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "eqWord8#"
+
+eqWord16HashValName :: Name
+eqWord16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "eqWord16#"
+
+extendInt8HashValName :: Name
+extendInt8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "extendInt8#"
+
+extendInt16HashValName :: Name
+extendInt16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "extendInt16#"
+
+extendWord8HashValName :: Name
+extendWord8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "extendWord8#"
+
+extendWord16HashValName :: Name
+extendWord16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "extendWord16#"
+
+geInt8HashValName :: Name
+geInt8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "geInt8#"
+
+geInt16HashValName :: Name
+geInt16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "geInt16#"
+
+geWord8HashValName :: Name
+geWord8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "geWord8#"
+
+geWord16HashValName :: Name
+geWord16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "geWord16#"
+
+gtInt8HashValName :: Name
+gtInt8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "gtInt8#"
+
+gtInt16HashValName :: Name
+gtInt16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "gtInt16#"
+
+gtWord8HashValName :: Name
+gtWord8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "gtWord8#"
+
+gtWord16HashValName :: Name
+gtWord16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "gtWord16#"
+
+int8HashTypeName :: Name
+int8HashTypeName = mkNameG_tc "ghc-prim" "GHC.Prim" "Int8#"
+
+int16HashTypeName :: Name
+int16HashTypeName = mkNameG_tc "ghc-prim" "GHC.Prim" "Int16#"
+
+leInt8HashValName :: Name
+leInt8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "leInt8#"
+
+leInt16HashValName :: Name
+leInt16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "leInt16#"
+
+leWord8HashValName :: Name
+leWord8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "leWord8#"
+
+leWord16HashValName :: Name
+leWord16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "leWord16#"
+
+ltInt8HashValName :: Name
+ltInt8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "ltInt8#"
+
+ltInt16HashValName :: Name
+ltInt16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "ltInt16#"
+
+ltWord8HashValName :: Name
+ltWord8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "ltWord8#"
+
+ltWord16HashValName :: Name
+ltWord16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "ltWord16#"
+
+word8HashTypeName :: Name
+word8HashTypeName = mkNameG_tc "ghc-prim" "GHC.Prim" "Word8#"
+
+word16HashTypeName :: Name
+word16HashTypeName = mkNameG_tc "ghc-prim" "GHC.Prim" "Word16#"
+#endif
