@@ -13,6 +13,9 @@ exports functionality which emulates the @GeneralizedNewtypeDeriving@ and
 
 On older versions of @template-haskell@/GHC, this module does not export
 anything.
+
+Note: this is an internal module, and as such, the API presented here is not
+guaranteed to be stable, even between minor releases of this library.
 -}
 module Data.Deriving.Via.Internal where
 
@@ -106,9 +109,14 @@ deriveViaDecs instanceTy mbViaTy = do
                   Nothing ->
                     case dataTy of
                       ConT dataName -> do
-                        DatatypeInfo { datatypeVars    = dataVars
-                                     , datatypeVariant = dv
-                                     , datatypeCons    = cons
+                        DatatypeInfo {
+#if MIN_VERSION_th_abstraction(0,3,0)
+                                       datatypeInstTypes = dataInstTypes
+#else
+                                       datatypeVars      = dataInstTypes
+#endif
+                                     , datatypeVariant   = dv
+                                     , datatypeCons      = cons
                                      } <- reifyDatatype dataName
                         case newtypeRepType dv cons of
                           Just newtypeRepTy ->
@@ -117,7 +125,7 @@ deriveViaDecs instanceTy mbViaTy = do
                                 let repTySubst =
                                       M.fromList $
                                       zipWith (\var arg -> (varTToName var, arg))
-                                              dataVars dataArgs
+                                              dataInstTypes dataArgs
                                 in return $ applySubstitution repTySubst etaRepTy
                               Nothing -> etaReductionError instanceTy
                           Nothing -> fail $ "Not a newtype: " ++ nameBase dataName
@@ -139,15 +147,19 @@ deriveViaDecs' clsName clsTvbs clsArgs repTy dec = do
   where
     go :: Dec -> Q (Maybe [Dec])
 
-    go (OpenTypeFamilyD (TypeFamilyHead tfName tfTvbs _ _)) =
+    go (OpenTypeFamilyD (TypeFamilyHead tfName tfTvbs _ _)) = do
       let lhsSubst = zipTvbSubst clsTvbs clsArgs
           rhsSubst = zipTvbSubst clsTvbs $ changeLast clsArgs repTy
           tfTvbTys = map tvbToType tfTvbs
           tfLHSTys = map (applySubstitution lhsSubst) tfTvbTys
           tfRHSTys = map (applySubstitution rhsSubst) tfTvbTys
           tfRHSTy  = applyTy (ConT tfName) tfRHSTys
-          tfInst   = TySynInstD tfName (TySynEqn tfLHSTys tfRHSTy)
-      in return (Just [tfInst])
+      tfInst <- tySynInstDCompat tfName
+#if MIN_VERSION_th_abstraction(0,3,0)
+                                 Nothing
+#endif
+                                 (map pure tfLHSTys) (pure tfRHSTy)
+      pure (Just [tfInst])
 
     go (SigD methName methTy) =
       let (fromTy, toTy) = mkCoerceClassMethEqn clsTvbs clsArgs repTy $
@@ -184,7 +196,12 @@ changeLast [_]    x  = [x]
 changeLast (x:xs) x' = x : changeLast xs x'
 
 stripOuterForallT :: Type -> Type
+#if __GLASGOW_HASKELL__ < 807
+-- Before GHC 8.7, TH-reified classes would put a redundant forall/class
+-- context in front of each method's type signature, so we have to strip them
+-- off here.
 stripOuterForallT (ForallT _ _ ty) = ty
+#endif
 stripOuterForallT ty               = ty
 
 decomposeType :: Type -> (Cxt, Type)
