@@ -9,6 +9,9 @@ Maintainer:  Ryan Scott
 Portability: Template Haskell
 
 Exports functions to mechanically derive 'Ord', 'Ord1', and 'Ord2' instances.
+
+Note: this is an internal module, and as such, the API presented here is not
+guaranteed to be stable, even between minor releases of this library.
 -}
 module Data.Ord.Deriving.Internal (
       -- * 'Ord'
@@ -37,7 +40,6 @@ module Data.Ord.Deriving.Internal (
 import           Data.Deriving.Internal
 import           Data.List (partition)
 import qualified Data.Map as Map
-import           Data.Maybe (isJust)
 
 import           Language.Haskell.TH.Datatype
 import           Language.Haskell.TH.Lib
@@ -160,23 +162,27 @@ deriveOrdClass :: OrdClass -> Name -> Q [Dec]
 deriveOrdClass oClass name = do
   info <- reifyDatatype name
   case info of
-    DatatypeInfo { datatypeContext = ctxt
-                 , datatypeName    = parentName
-                 , datatypeVars    = vars
-                 , datatypeVariant = variant
-                 , datatypeCons    = cons
+    DatatypeInfo { datatypeContext   = ctxt
+                 , datatypeName      = parentName
+#if MIN_VERSION_th_abstraction(0,3,0)
+                 , datatypeInstTypes = instTypes
+#else
+                 , datatypeVars      = instTypes
+#endif
+                 , datatypeVariant   = variant
+                 , datatypeCons      = cons
                  } -> do
       (instanceCxt, instanceType)
-          <- buildTypeInstance oClass parentName ctxt vars variant
+          <- buildTypeInstance oClass parentName ctxt instTypes variant
       (:[]) `fmap` instanceD (return instanceCxt)
                              (return instanceType)
-                             (ordFunDecs oClass vars cons)
+                             (ordFunDecs oClass instTypes cons)
 
 -- | Generates a declaration defining the primary function(s) corresponding to a
 -- particular class (compare for Ord, liftCompare for Ord1, and
 -- liftCompare2 for Ord2).
 ordFunDecs :: OrdClass -> [Type] -> [ConstructorInfo] -> [Q Dec]
-ordFunDecs oClass vars cons =
+ordFunDecs oClass instTypes cons =
     map makeFunD $ ordClassToCompare oClass : otherFuns oClass cons
   where
     makeFunD :: OrdFun -> Q Dec
@@ -206,7 +212,7 @@ ordFunDecs oClass vars cons =
                                    , Ord1Compare1
 #endif
                                    ]
-                      = makeOrdFunForCons oFun vars cons
+                      = makeOrdFunForCons oFun instTypes cons
     dispatchFun OrdLE = dispatchLT $ \lt x y -> negateExpr $ lt `appE` y `appE` x
     dispatchFun OrdGT = dispatchLT $ \lt x y ->              lt `appE` y `appE` x
     dispatchFun OrdGE = dispatchLT $ \lt x y -> negateExpr $ lt `appE` x `appE` y
@@ -219,25 +225,29 @@ makeOrdFun :: OrdFun -> [Q Match] -> Name -> Q Exp
 makeOrdFun oFun matches name = do
   info <- reifyDatatype name
   case info of
-    DatatypeInfo { datatypeContext = ctxt
-                 , datatypeName    = parentName
-                 , datatypeVars    = vars
-                 , datatypeVariant = variant
-                 , datatypeCons    = cons
+    DatatypeInfo { datatypeContext   = ctxt
+                 , datatypeName      = parentName
+#if MIN_VERSION_th_abstraction(0,3,0)
+                 , datatypeInstTypes = instTypes
+#else
+                 , datatypeVars      = instTypes
+#endif
+                 , datatypeVariant   = variant
+                 , datatypeCons      = cons
                  } -> do
       let oClass = ordFunToClass oFun
           others = otherFuns oClass cons
       -- We force buildTypeInstance here since it performs some checks for whether
       -- or not the provided datatype can actually have compare/liftCompare/etc.
       -- implemented for it, and produces errors if it can't.
-      buildTypeInstance oClass parentName ctxt vars variant >>
+      buildTypeInstance oClass parentName ctxt instTypes variant >>
         if oFun `elem` compareFuns || oFun `elem` others
-           then makeOrdFunForCons oFun vars cons
+           then makeOrdFunForCons oFun instTypes cons
            else do
              x <- newName "x"
              y <- newName "y"
              lamE [varP x, varP y] $
-                  caseE (makeOrdFunForCons (ordClassToCompare oClass) vars cons
+                  caseE (makeOrdFunForCons (ordClassToCompare oClass) instTypes cons
                              `appE` varE x `appE` varE y)
                         matches
   where
@@ -254,7 +264,7 @@ makeOrdFun oFun matches name = do
 -- | Generates a lambda expression for the given constructors.
 -- All constructors must be from the same type.
 makeOrdFunForCons :: OrdFun -> [Type] -> [ConstructorInfo] -> Q Exp
-makeOrdFunForCons oFun vars cons = do
+makeOrdFunForCons oFun instTypes cons = do
     let oClass = ordFunToClass oFun
     v1     <- newName "v1"
     v2     <- newName "v2"
@@ -263,7 +273,7 @@ makeOrdFunForCons oFun vars cons = do
     ords   <- newNameList "ord" $ arity oClass
 
     let lastTyVars :: [Name]
-        lastTyVars = map varTToName $ drop (length vars - fromEnum oClass) vars
+        lastTyVars = map varTToName $ drop (length instTypes - fromEnum oClass) instTypes
 
         tvMap :: TyVarMap1
         tvMap = Map.fromList $ zipWith (\x y -> (x, OneName y)) lastTyVars ords
@@ -677,52 +687,13 @@ unliftedCompare ltFun eqFun aExpr bExpr lt eq gt =
     ascribeBool e = sigE e $ conT boolTypeName
 
 primOrdFuns :: Name -> (Name, Name, Name, Name, Name)
-primOrdFuns tyName = case lookup tyName primOrdFunTbl of
+primOrdFuns tyName =
+  case Map.lookup tyName primOrdFunTbl of
     Just names -> names
     Nothing    -> error $ nameBase tyName ++ " is not supported."
 
-primOrdFunTbl :: [(Name, (Name, Name, Name, Name, Name))]
-primOrdFunTbl =
-    [ (addrHashTypeName,   ( ltAddrHashValName
-                           , leAddrHashValName
-                           , eqAddrHashValName
-                           , geAddrHashValName
-                           , gtAddrHashValName
-                           ))
-    , (charHashTypeName,   ( ltCharHashValName
-                           , leCharHashValName
-                           , eqCharHashValName
-                           , geCharHashValName
-                           , gtCharHashValName
-                           ))
-    , (doubleHashTypeName, ( ltDoubleHashValName
-                           , leDoubleHashValName
-                           , eqDoubleHashValName
-                           , geDoubleHashValName
-                           , gtDoubleHashValName
-                           ))
-    , (floatHashTypeName,  ( ltFloatHashValName
-                           , leFloatHashValName
-                           , eqFloatHashValName
-                           , geFloatHashValName
-                           , gtFloatHashValName
-                           ))
-    , (intHashTypeName,    ( ltIntHashValName
-                           , leIntHashValName
-                           , eqIntHashValName
-                           , geIntHashValName
-                           , gtIntHashValName
-                           ))
-    , (wordHashTypeName,   ( ltWordHashValName
-                           , leWordHashValName
-                           , eqWordHashValName
-                           , geWordHashValName
-                           , gtWordHashValName
-                           ))
-    ]
-
 isSupportedUnliftedType :: Type -> Bool
-isSupportedUnliftedType (ConT tyName) = isJust $ lookup tyName primOrdFunTbl
+isSupportedUnliftedType (ConT tyName) = Map.member tyName primOrdFunTbl
 isSupportedUnliftedType _             = False
 
 isSingleton :: [a] -> Bool
