@@ -1,5 +1,6 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE GADTs #-}
+{-# LANGUAGE NamedFieldPuns #-}
 {-|
 Module:      Text.Show.Deriving.Internal
 Copyright:   (C) 2015-2017 Ryan Scott
@@ -8,6 +9,9 @@ Maintainer:  Ryan Scott
 Portability: Template Haskell
 
 Exports functions to mechanically derive 'Show', 'Show1', and 'Show2' instances.
+
+Note: this is an internal module, and as such, the API presented here is not
+guaranteed to be stable, even between minor releases of this library.
 -}
 module Text.Show.Deriving.Internal (
       -- * 'Show'
@@ -50,6 +54,7 @@ module Text.Show.Deriving.Internal (
 import           Data.Deriving.Internal
 import           Data.List
 import qualified Data.Map as Map
+import           Data.Map (Map)
 import           Data.Maybe (fromMaybe)
 
 import           GHC.Show (appPrec, appPrec1)
@@ -268,26 +273,30 @@ deriveShowClass :: ShowClass -> ShowOptions -> Name -> Q [Dec]
 deriveShowClass sClass opts name = do
   info <- reifyDatatype name
   case info of
-    DatatypeInfo { datatypeContext = ctxt
-                 , datatypeName    = parentName
-                 , datatypeVars    = vars
-                 , datatypeVariant = variant
-                 , datatypeCons    = cons
+    DatatypeInfo { datatypeContext   = ctxt
+                 , datatypeName      = parentName
+#if MIN_VERSION_th_abstraction(0,3,0)
+                 , datatypeInstTypes = instTypes
+#else
+                 , datatypeVars      = instTypes
+#endif
+                 , datatypeVariant   = variant
+                 , datatypeCons      = cons
                  } -> do
       (instanceCxt, instanceType)
-          <- buildTypeInstance sClass parentName ctxt vars variant
+          <- buildTypeInstance sClass parentName ctxt instTypes variant
       (:[]) `fmap` instanceD (return instanceCxt)
                              (return instanceType)
-                             (showsPrecDecs sClass opts vars cons)
+                             (showsPrecDecs sClass opts instTypes cons)
 
 -- | Generates a declaration defining the primary function corresponding to a
 -- particular class (showsPrec for Show, liftShowsPrec for Show1, and
 -- liftShowsPrec2 for Show2).
 showsPrecDecs :: ShowClass -> ShowOptions -> [Type] -> [ConstructorInfo] -> [Q Dec]
-showsPrecDecs sClass opts vars cons =
+showsPrecDecs sClass opts instTypes cons =
     [ funD (showsPrecName sClass)
            [ clause []
-                    (normalB $ makeShowForCons sClass opts vars cons)
+                    (normalB $ makeShowForCons sClass opts instTypes cons)
                     []
            ]
     ]
@@ -298,29 +307,33 @@ makeShowsPrecClass :: ShowClass -> ShowOptions -> Name -> Q Exp
 makeShowsPrecClass sClass opts name = do
   info <- reifyDatatype name
   case info of
-    DatatypeInfo { datatypeContext = ctxt
-                 , datatypeName    = parentName
-                 , datatypeVars    = vars
-                 , datatypeVariant = variant
-                 , datatypeCons    = cons
+    DatatypeInfo { datatypeContext   = ctxt
+                 , datatypeName      = parentName
+#if MIN_VERSION_th_abstraction(0,3,0)
+                 , datatypeInstTypes = instTypes
+#else
+                 , datatypeVars      = instTypes
+#endif
+                 , datatypeVariant   = variant
+                 , datatypeCons      = cons
                  } -> do
       -- We force buildTypeInstance here since it performs some checks for whether
       -- or not the provided datatype can actually have showsPrec/liftShowsPrec/etc.
       -- implemented for it, and produces errors if it can't.
-      buildTypeInstance sClass parentName ctxt vars variant
-        >> makeShowForCons sClass opts vars cons
+      buildTypeInstance sClass parentName ctxt instTypes variant
+        >> makeShowForCons sClass opts instTypes cons
 
 -- | Generates a lambda expression for showsPrec/liftShowsPrec/etc. for the
 -- given constructors. All constructors must be from the same type.
 makeShowForCons :: ShowClass -> ShowOptions -> [Type] -> [ConstructorInfo] -> Q Exp
-makeShowForCons sClass opts vars cons = do
+makeShowForCons sClass opts instTypes cons = do
     p     <- newName "p"
     value <- newName "value"
     sps   <- newNameList "sp" $ arity sClass
     sls   <- newNameList "sl" $ arity sClass
     let spls       = zip sps sls
         _spsAndSls = interleave sps sls
-        lastTyVars = map varTToName $ drop (length vars - fromEnum sClass) vars
+        lastTyVars = map varTToName $ drop (length instTypes - fromEnum sClass) instTypes
         splMap     = Map.fromList $ zipWith (\x (y, z) -> (x, TwoNames y z)) lastTyVars spls
 
         makeFun
@@ -490,27 +503,24 @@ makeShowForArg p _ opts _ _ (ConT tyName) tyExpName =
     tyVarE = varE tyExpName
 
     showE :: Q Exp
-    showE | tyName == charHashTypeName   = showPrimE cHashDataName oneHashE
-          | tyName == doubleHashTypeName = showPrimE dHashDataName twoHashE
-          | tyName == floatHashTypeName  = showPrimE fHashDataName oneHashE
-          | tyName == intHashTypeName    = showPrimE iHashDataName oneHashE
-          | tyName == wordHashTypeName   = showPrimE wHashDataName twoHashE
-          | otherwise = varE showsPrecValName `appE` integerE p `appE` tyVarE
+    showE =
+      case Map.lookup tyName primShowTbl of
+        Just ps -> showPrimE ps
+        Nothing -> varE showsPrecValName `appE` integerE p `appE` tyVarE
 
-    -- Starting with GHC 7.10, data types containing unlifted types with derived Show
-    -- instances show hashed literals with actual hash signs, and negative hashed
-    -- literals are not surrounded with parentheses.
-    showPrimE :: Name -> Q Exp -> Q Exp
-    showPrimE con hashE
+    showPrimE :: PrimShow -> Q Exp
+    showPrimE PrimShow{primShowBoxer, primShowPostfixMod, primShowConv}
+        -- Starting with GHC 8.0, data types containing unlifted types with
+        -- derived Show instances show hashed literals with actual hash signs,
+        -- and negative hashed literals are not surrounded with parentheses.
       | ghc8ShowBehavior opts
-      = infixApp (varE showsPrecValName `appE` integerE 0 `appE` (conE con `appE` tyVarE))
-                 (varE composeValName)
-                 hashE
-      | otherwise = varE showsPrecValName `appE` integerE p `appE` (conE con `appE` tyVarE)
-
-    oneHashE, twoHashE :: Q Exp
-    oneHashE = varE showCharValName `appE` charE '#'
-    twoHashE = varE showStringValName `appE` stringE "##"
+      = primShowConv $ infixApp (primE 0) (varE composeValName) primShowPostfixMod
+      | otherwise
+      = primE p
+      where
+        primE :: Int -> Q Exp
+        primE prec = varE showsPrecValName `appE` integerE prec
+                                           `appE` primShowBoxer tyVarE
 makeShowForArg p sClass _ conName tvMap ty tyExpName =
     makeShowForType sClass conName tvMap False ty `appE` integerE p `appE` varE tyExpName
 
@@ -654,3 +664,74 @@ parenInfixConName conName =
 
 charE :: Char -> Q Exp
 charE = litE . charL
+
+data PrimShow = PrimShow
+  { primShowBoxer      :: Q Exp -> Q Exp
+  , primShowPostfixMod :: Q Exp
+  , primShowConv       :: Q Exp -> Q Exp
+  }
+
+primShowTbl :: Map Name PrimShow
+primShowTbl = Map.fromList
+    [ (charHashTypeName,   PrimShow
+                             { primShowBoxer      = appE (conE cHashDataName)
+                             , primShowPostfixMod = oneHashE
+                             , primShowConv       = id
+                             })
+    , (doubleHashTypeName, PrimShow
+                             { primShowBoxer      = appE (conE dHashDataName)
+                             , primShowPostfixMod = twoHashE
+                             , primShowConv       = id
+                             })
+    , (floatHashTypeName,  PrimShow
+                             { primShowBoxer      = appE (conE fHashDataName)
+                             , primShowPostfixMod = oneHashE
+                             , primShowConv       = id
+                             })
+    , (intHashTypeName,    PrimShow
+                             { primShowBoxer      = appE (conE iHashDataName)
+                             , primShowPostfixMod = oneHashE
+                             , primShowConv       = id
+                             })
+    , (wordHashTypeName,   PrimShow
+                             { primShowBoxer      = appE (conE wHashDataName)
+                             , primShowPostfixMod = twoHashE
+                             , primShowConv       = id
+                             })
+#if MIN_VERSION_base(4,13,0)
+    , (int8HashTypeName,   PrimShow
+                             { primShowBoxer      = appE (conE iHashDataName) . appE (varE extendInt8HashValName)
+                             , primShowPostfixMod = oneHashE
+                             , primShowConv       = mkNarrowE "narrowInt8#"
+                             })
+    , (int16HashTypeName,  PrimShow
+                             { primShowBoxer      = appE (conE iHashDataName) . appE (varE extendInt16HashValName)
+                             , primShowPostfixMod = oneHashE
+                             , primShowConv       = mkNarrowE "narrowInt16#"
+                             })
+    , (word8HashTypeName,  PrimShow
+                             { primShowBoxer      = appE (conE wHashDataName) . appE (varE extendWord8HashValName)
+                             , primShowPostfixMod = twoHashE
+                             , primShowConv       = mkNarrowE "narrowWord8#"
+                             })
+    , (word16HashTypeName, PrimShow
+                             { primShowBoxer      = appE (conE wHashDataName) . appE (varE extendWord16HashValName)
+                             , primShowPostfixMod = twoHashE
+                             , primShowConv       = mkNarrowE "narrowWord16#"
+                             })
+#endif
+    ]
+
+#if MIN_VERSION_base(4,13,0)
+mkNarrowE :: String -> Q Exp -> Q Exp
+mkNarrowE narrowStr e =
+  foldr (`infixApp` varE composeValName)
+        (varE showCharValName `appE` charE ')')
+        [ varE showStringValName `appE` stringE ('(':narrowStr ++ " ")
+        , e
+        ]
+#endif
+
+oneHashE, twoHashE :: Q Exp
+oneHashE = varE showCharValName `appE` charE '#'
+twoHashE = varE showStringValName `appE` stringE "##"
