@@ -9,6 +9,11 @@
 {-# LANGUAGE TemplateHaskell #-}
 {-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE UndecidableInstances #-}
+#if __GLASGOW_HASKELL__ >= 708
+{-# LANGUAGE EmptyCase #-}
+{-# LANGUAGE RoleAnnotations #-}
+#endif
+
 {-# OPTIONS_GHC -fno-warn-name-shadowing #-}
 {-# OPTIONS_GHC -fno-warn-unused-matches #-}
 #if __GLASGOW_HASKELL__ >= 800
@@ -27,8 +32,8 @@ Portability: Template Haskell
 module FunctorSpec where
 
 import Data.Char (chr)
-import Data.Deriving
 import Data.Foldable (fold)
+import Data.Deriving
 import Data.Functor.Classes (Eq1)
 import Data.Functor.Compose (Compose(..))
 import Data.Functor.Identity (Identity(..))
@@ -101,6 +106,12 @@ data IntHash a b
 data IntHashFun a b
     = IntHashFun ((((a -> Int#) -> b) -> Int#) -> a)
 
+data Empty1 a
+data Empty2 a
+#if __GLASGOW_HASKELL__ >= 708
+type role Empty2 nominal
+#endif
+
 -- Data families
 
 data family   StrangeFam x  y z
@@ -207,6 +218,15 @@ $(deriveTraversable ''IntHash)
 
 $(deriveFunctor     ''IntHashFun)
 
+$(deriveFunctor     ''Empty1)
+$(deriveFoldable    ''Empty1)
+$(deriveTraversable ''Empty1)
+
+-- Use EmptyCase here
+$(deriveFunctorOptions     defaultFFTOptions{ fftEmptyCaseBehavior = True } ''Empty2)
+$(deriveFoldableOptions    defaultFFTOptions{ fftEmptyCaseBehavior = True } ''Empty2)
+$(deriveTraversableOptions defaultFFTOptions{ fftEmptyCaseBehavior = True } ''Empty2)
+
 #if MIN_VERSION_template_haskell(2,7,0)
 -- Data families
 
